// RegKey.cpp: implementation of the CRegKey class.
//

// This file is part of the Xeyes for windows program (c) copyright 1998 Shachar Shemesh
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation version 2 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//                                             
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RegKey.h"

//////////////////////////////////////////////////////////////////////
// Global constants
//////////////////////////////////////////////////////////////////////

const CRegKey HKCR( HKEY_CLASSES_ROOT );
const CRegKey HKCU( HKEY_CURRENT_USER );
const CRegKey HKLM( HKEY_LOCAL_MACHINE );
const CRegKey HKU( HKEY_USERS );
const CRegKey HKPD( HKEY_PERFORMANCE_DATA );
const CRegKey HKCC( HKEY_CURRENT_CONFIG );
const CRegKey HKDD( HKEY_DYN_DATA );

//////////////////////////////////////////////////////////////////////
// Static definitions
//////////////////////////////////////////////////////////////////////
CRegKey::keyData CRegKey::KeyData;

//////////////////////////////////////////////////////////////////////
// Member functions
//////////////////////////////////////////////////////////////////////

#ifdef REGKEY_USE_DUP
HKEY CRegKey::DupHandle( HKEY hKey )
{
   if( ((DWORD)hKey ^ (DWORD)HKEY_CLASSES_ROOT) < 16 ) // Do not duplicate predefined handles
      return hKey;

   HKEY newHandle;
   if( !DuplicateHandle( GetCurrentProcess(), hKey, GetCurrentProcess(),
      (LPHANDLE)&newHandle, 0, TRUE, DUPLICATE_SAME_ACCESS ) )
      throw Except( GetLastError() );
      
   return newHandle;
}
#endif

void CRegKey::GetKeyInfo() const
// Updates the KeyData structure for this key
{
   if( KeyData.hKey!=m_hKey )
   {
      KeyData.hKey=NULL;
      LONG error=RegQueryInfoKey( m_hKey, NULL, NULL, NULL, &KeyData.nSubKeys,
         &KeyData.nMaxSubkeyLen, NULL, &KeyData.nValues, &KeyData.nMaxValueName,
         &KeyData.nMaxValueLen, &KeyData.SecurityDesc, NULL );

      if( error!=ERROR_SUCCESS )
      {
         throw Except( error );
      }

      KeyData.hKey=m_hKey;
   }
}

CRegKey CRegKey::Create( const char * name, REGSAM Access, DWORD dwOptions ) const
{
   KeyData.hKey=NULL;
   HKEY hkTemp;
   CRegKey retval;
   DWORD created;

   LONG result=RegCreateKeyEx( m_hKey, name, 0, NULL, dwOptions, Access, NULL, &hkTemp,
      &created );
   if( result==ERROR_SUCCESS )
      retval=CRegKey( hkTemp );
   else
      throw Except( result );

   return retval;
}

CRegKey CRegKey::Open( const char * name, REGSAM Access ) const
{
   HKEY hkTemp;
   CRegKey retval;

   LONG result=RegOpenKeyEx( m_hKey, name, 0, Access, &hkTemp );
   if( result==ERROR_SUCCESS )
      retval=CRegKey( hkTemp );
   else
      throw Except( result );

   return retval;
}

string CRegKey::EnumSubkey(DWORD index) const
{
   const MAXPATH=260;
   DWORD buffersize=MAXPATH+1; // RegQueryInfoKey is unreliable with max key length
   auto_array<char> buffer( new char [ buffersize ] );

   LONG result=RegEnumKeyEx( m_hKey, index, buffer, &buffersize, NULL, NULL, NULL,
      NULL );
   if( result!=ERROR_SUCCESS )
   {
      throw Except( result );
   }

   return string( buffer );
}

string CRegKey::EnumValue(DWORD index)
{
   DWORD buffersize=QueryMaxValueName()+1;
   auto_array<char> buffer( new char [ buffersize ] );

   LONG result=RegEnumValue( m_hKey, index, buffer, &buffersize, NULL, NULL, NULL,
      NULL );
   if( result!=ERROR_SUCCESS )
   {
      throw Except( result );
   }

   return string( buffer );
}

void CRegKey::DeleteKey( const char * subKey )
{
   try {
      CRegKey current=Open( subKey );
      
      for( DWORD nSubKeys=current.QueryNumSubkeys(); nSubKeys>0; --nSubKeys )
      {
         current.DeleteKey( current.EnumSubkey( nSubKeys-1 ).c_str() );
      }
   } // Remove current from context - close the key.
   catch( const Except &exc )
   {
      // Make sure we don't throw "File not found" on non-existing keys we want to delete (stupid)
      if( exc.get_error()!=ERROR_FILE_NOT_FOUND && exc.get_error()!=ERROR_PATH_NOT_FOUND )
         throw;
   }

   LONG result=RegDeleteKey( m_hKey, subKey );

   if( result!=ERROR_SUCCESS && result!=ERROR_FILE_NOT_FOUND && result!=ERROR_PATH_NOT_FOUND )
   {
      throw Except( result );
   }
}

void CRegKey::DeleteValue( const char * value )
{
   LONG result=RegDeleteValue( m_hKey, value );
   if( result!=ERROR_SUCCESS && result!=ERROR_FILE_NOT_FOUND && result!=ERROR_PATH_NOT_FOUND )
      throw Except( result );
}

void CRegKey::SetValue( const char * value, const char * data )
{ // The REG_STR version
   LONG result=RegSetValueEx( m_hKey, value, 0, REG_SZ, (LPBYTE)data, strlen( data )+1 );
   if( result!=ERROR_SUCCESS )
      throw Except( result );
}

void CRegKey::SetValue( const char * value, DWORD data )
{ // The REG_DWORD version
   LONG result=RegSetValueEx( m_hKey, value, 0, REG_DWORD, (LPBYTE)&data, sizeof( data ) );
   if( result!=ERROR_SUCCESS )
      throw Except( result );
}

void CRegKey::SetValue( const char * value, const std::basic_string< unsigned char > & data,
              DWORD type )
{
   LONG result=RegSetValueEx( m_hKey, value, 0, type, data.data(), data.size() );
   if( result!=ERROR_SUCCESS )
      throw Except( result );
}

void CRegKey::SetValue( const char * value, const std::vector< string > & data )
{
   std::basic_string< unsigned char > write_data;

#ifndef UNICODE
   

   for( std::vector< string >::const_iterator i=data.begin(); i!=data.end(); ++i )
   {
      for( string::const_iterator j=i->begin(); j!=i->end(); ++j )
      {
         write_data+=(unsigned char)*j;
      }

      write_data+=(unsigned char)'\0';
   }
   write_data+=(unsigned char)'\0';

   SetValue( value, write_data, REG_MULTI_SZ );
#else
#error Need to make sure above code works with UNICODE
#endif
}

DWORD CRegKey::QueryValueType( const char * value ) const
{ // This function returns REG_NONE if the value doesn't exist. It will not throw an exception in that case
   DWORD type=REG_NONE;
   LONG result=RegQueryValueEx( m_hKey, value, NULL, &type, NULL, NULL );
   if( result==ERROR_FILE_NOT_FOUND )
   {
      type=REG_NONE;
      result=ERROR_SUCCESS;
   }

   if( result!=ERROR_SUCCESS )
   {
      throw Except( result );
   }

   return type;
}

DWORD CRegKey::QueryValue( const char * name, DWORD Default ) const
{
   DWORD type=QueryValueType( name );

   if( type!=REG_NONE && type!=REG_DWORD )
      throw Except( ERROR_INVALID_FUNCTION ); // Attempt to treat non-DW as DW

   DWORD data=Default;
   DWORD size=sizeof( DWORD );
   if( type==REG_DWORD )
   {
      LONG result=RegQueryValueEx( m_hKey, name, NULL, NULL, (LPBYTE)&data, &size );
      if( result!=ERROR_SUCCESS )
         throw Except( result );
   }

   return data;
}

string CRegKey::QueryValue( const char * name, const string &Default ) const
{
   DWORD type=QueryValueType( name );

   if( type!=REG_NONE && type!=REG_SZ )
      throw Except( ERROR_INVALID_FUNCTION ); // Attempt to treat non-DW as DW

   DWORD buffersize=QueryMaxValueLen();
   auto_array<char> buffer( new char [ buffersize ] );
   if( type==REG_SZ )
   {
      LONG result=RegQueryValueEx( m_hKey, name, NULL, NULL, (LPBYTE)static_cast<char *>(buffer), &buffersize );
      if( result!=ERROR_SUCCESS )
         throw Except( result );
   }

   return type==REG_SZ ? string( buffer, buffersize-1 ) : Default;
}

std::basic_string< unsigned char > CRegKey::QueryValue( const char * name ) const
{
   DWORD buffersize=QueryMaxValueLen();
   auto_array<unsigned char> buffer( new unsigned char [ buffersize ] );
   LONG result=RegQueryValueEx( m_hKey, name, NULL, NULL, buffer, &buffersize );
   if( result!=ERROR_SUCCESS )
      throw Except( result );

   return std::basic_string<unsigned char>( buffer, buffersize );
}

std::vector< string > CRegKey::QueryValue( const char * name, const std::vector< string > &Default ) const // Used for MultiSZ query
{
   DWORD type=QueryValueType( name );
   if(  type!=REG_MULTI_SZ && type!=REG_NONE )
      throw Except( ERROR_INVALID_FUNCTION ); // Attempt to treat a non REG_MULTI_SZ as such

   std::vector< string > retval;

   if( type==REG_MULTI_SZ )
   {
      std::basic_string< unsigned char > data=QueryValue( name );

      string temp;
      for( std::basic_string< unsigned char >::const_iterator i=data.begin(); i!=data.end();
         ++i )
      {
         if( *i!='\0' )
         {
            temp+=(char)*i;
         } else
         {
            if( !temp.empty() )
            {
               retval.push_back( temp );
               temp.resize( 0 );
            }
         }
      }
   } else
      retval=Default;

   return retval;
}
